"""
A rich-feature JSON parser module - support Python classes.
"""

__version__ = "v1.0.0"

try:
    from ujson import JSONDecodeError, loads as _json_loads, dumps as _json_dumps
except ImportError:
    raise ImportError(
        "ujsonx requires UltraJSON module. install it by 'pip3 install ujson'."
    ) from None

import typing
import warnings

class DataIsNullError(Exception):
    """
    Data is null error.
    """
    pass

class _SubscriptableType:
    """
    Create subscriptable type!
    """
    def _getter(*args):
        raise NotImplementedError
    
    def __class_getitem__(cls, params):
        if isinstance(params, tuple):
            return cls._getter(*params)
        
        return cls._getter(params)

class _JBase:
    pass

class JTag(_SubscriptableType, _JBase):
    """
    Specify your class field type.
    """
    def __init__(
        self, init: typing.Optional[typing.Callable] = None, name: str = "",
        omitempty: bool = False, _default=None, ignore: bool = False
    ) -> None:
        self.init = init
        self.name = name
        self.omitempty = omitempty
        self._default = _default
        self.ignore = ignore
    
    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(init={self.init}, name={self.name}, omitempty={self.omitempty}," \
            f" _default={self._default}, ignore={self.ignore})"

    def _getter(*args):
        return JTag(*args)

class JNested(_SubscriptableType, _JBase):
    """
    Specify your class field type as nested field.
    """
    def __init__(
        self, init: type, name: str = "",
        omitempty: bool = False, _default=None
    ) -> None:
        self.init = init
        self.name = name
        self.omitempty = omitempty
        self._default = _default
    
    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(init={self.init}, name={self.name}, omitempty={self.omitempty}," \
            f" _default={self._default})"
    
    def _getter(*args):
        return JNested(*args)

class JList(_SubscriptableType, _JBase):
    def __init__(
        self, init: type, name: str = "", omitempty: bool = False, _default=None
    ) -> None:
        self.init = init
        self.name = name
        self.omitempty = omitempty
        self._default = _default
    
    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(init={self.init}, omitempty={self.omitempty}," \
            f" _default={self._default})"
    
    def _getter(*args):
        return JList(*args)

_RES_TYPE = typing.TypeVar("_RES_TYPE")
def _initialize_type(_type: typing.Optional[typing.Type[_RES_TYPE]]) -> _RES_TYPE:
    _old_init = _type.__init__

    def _empty_initialize(_):
        pass

    _type.__init__ = _empty_initialize
    result = _type()
    _type.__init__ = _old_init
    return result

def _load_to_type(
    data: typing.Any, _type: typing.Union[typing.Type[_RES_TYPE], JList, None], strict: bool, _should_list: bool
) -> typing.Union[_RES_TYPE, typing.List[_RES_TYPE]]:
    """
    use `loads` function instead of this.
    """
    if _type is None:
        return data

    IS_LIST = isinstance(data, list)

    if _should_list and not IS_LIST:
        raise TypeError(
            f"cannot loads {type(data).__name__} into value of type (list of) {_type.__name__}s"
        )

    if IS_LIST:
        if isinstance(_type, list):
            return list(
                _load_to_type(a, b, strict, False) for a, b in zip(data, _type)
            )
        
        if isinstance(_type, JList):
            return list(
                _load_to_type(a, _type.init, strict, False) for a in data
            )

        if strict:
            raise TypeError(
                f"cannot loads list into value of type {_type.__name__} (strict enabled)"
            )
        
        return list(
            _load_to_type(a, _type, strict, False) for a in data
        )
    
    if data is None:
        raise DataIsNullError(
            "data is null"
        )
    
    if _type in (int, float, str, dict, list):
        if not (type(data) == _type):
            raise TypeError(
                f"cannot loads {type(data).__name__} into value of type {_type.__name__}"
            )
        
        return data

    if not isinstance(data, dict):
        if isinstance(data, (int, float, str)) and (type(data) == _type):
            return data
        
        raise TypeError(
            f"cannot loads {type(data).__name__} into value of type {_type.__name__}"
        )
    
    result = _initialize_type(_type)

    for f_name, f_type in _type.__annotations__.items():
        if isinstance(f_type, JTag):
            if f_type.ignore:
                continue

            try:
                setattr(result, f_name, _load_to_type(data[f_type.name or f_name], f_type.init, strict, False))
            except KeyError:
                setattr(result, f_name, f_type._default)
            except (ValueError, TypeError):
                raise TypeError(
                        f"cannot loads {type(data[f_type.name or f_name]).__name__} into value of type {f_type.init.__name__} (field {_type.__name__}.{f_name})"
                    ) from None
        
        elif isinstance(f_type, JNested):
            try:
                setattr(result, f_name, _load_to_type(data[f_type.name or f_name], f_type.init, strict, False))
            except KeyError:
                setattr(result, f_name, f_type._default)
            except DataIsNullError:
                setattr(result, f_name, None)

        elif isinstance(f_type, JList):
            try:
                setattr(result, f_name, _load_to_type(data[f_type.name or f_name], f_type.init, False, True))
            except KeyError:
                setattr(result, f_name, f_type._default)
            except DataIsNullError:
                setattr(result, f_name, None)

        else:
            try:
                setattr(result, f_name, _load_to_type(data[f_name], f_type, strict, False))
            except KeyError:
                setattr(result, f_name, None)
            except (ValueError, TypeError):
                raise TypeError(
                    f"cannot loads {type(data[f_name]).__name__} into value of type {f_type.__name__} (field {_type.__name__}.{f_name})"
                ) from None
    
    return result

def loads(
    s: str, _type: typing.Union[typing.Type[_RES_TYPE], JList, None] = None, strict: bool = False
) -> typing.Union[_RES_TYPE, typing.List[_RES_TYPE]]:
    """
    Loads `s` to your `_type` type.

    if loaded `s` is list, returns list of `_type` (set `strict` to disable it);

    you can set limit for list length by passing `list[type, type, ...]` as `_type`.
    how? see example::
    
        res = loads('[{'id': 1}, {'id': 2}]', [dict])
        print(res) # [{'id': 1}]

    ##### Notes:
    * if `_type` is dict or list type, just loads `s` and returns it.
    * don't use `tuple` (object/type) for `_type` (unsupported).
    """
    return _load_to_type(_json_loads(s), _type, strict, False)

class NotSpecified:
    pass

def _dump_to_dict(_obj: typing.Any) -> dict:
    """
    Converts object to dictionary.

    ##### Note:
    * `list` not supported in this function - see `dumps` function.
    """
    if isinstance(_obj, (dict, int, float, str)) or _obj is None:
        return _obj
    
    result = {}

    for f_name, f_type in type(_obj).__annotations__.items():
        if isinstance(f_type, JTag):
            if f_type.ignore:
                continue

            attr = getattr(_obj, f_name, NotSpecified)

            if not (attr is NotSpecified):
                result[f_type.name or f_name] = attr
            
            elif not f_type.omitempty:
                result[f_type.name or f_name] = None

        elif isinstance(f_type, JNested):
            attr = getattr(_obj, f_name, NotSpecified)
            
            if not (attr is NotSpecified):
                result[f_type.name or f_name] = _dump_to_dict(attr)
            
            elif not f_type.omitempty:
                result[f_type.name or f_name] = None
        
        elif isinstance(f_type, JList):
            attr = getattr(_obj, f_name, NotSpecified)
            
            if (not attr is NotSpecified) and attr is not None:
                if not isinstance(attr, list):
                    warnings.warn(
                        "pycjson.dumps: you are using non-list type for pycjson.JList field -" \
                        " if you want to set this field empty, use None or pycjson.NotSpecified - field skipped.",
                        RuntimeWarning,
                        stacklevel=3
                    )
                
                else:
                    result[f_type.name or f_name] = list(_dump_to_dict(i) for i in attr)
            
            elif not f_type.omitempty:
                result[f_type.name or f_name] = None
        
        else:
            attr = getattr(_obj, f_name, NotSpecified)

            if not (attr is NotSpecified):
                result[f_name] = attr
            else:
                result[f_name] = None

    return result

@typing.overload
def dumps(
    _obj: typing.Any, ensure_ascii: bool = ..., encode_html_chars: bool = ...,
    escape_forward_slashes: bool = ..., sort_keys: bool = ..., indent: int = ...
) -> str:
    """
    Converts any object to JSON string.

    Parameters:
        `ensure_ascii` - Limits output to ASCII and escapes all extended characters above 127. Default is `True`. If your end format supports UTF-8, setting this option to false is highly recommended to save space.

        `encode_html_chars` - Used to enable special encoding of "unsafe" HTML characters into safer Unicode sequences.Default is `False`.
    
        `escape_forward_slashes` - Controls whether forward slashes (/) are escaped. Default is `True`.

        `sort_keys` - if `True`, the output of dictionaries will be sorted by key. Default is `False`.

        `indent` - Controls whether indentation ("pretty output") is enabled. Default is `0` (disabled).
    """
    ...

def dumps(_obj: typing.Any, *args, **kwargs) -> str:    
    if isinstance(_obj, list):
        return _json_dumps(
            list(
                _dump_to_dict(i) for i in _obj
            ),
            *args,
            **kwargs
        )

    return _json_dumps(_dump_to_dict(_obj), *args, **kwargs)
